@@ -1,3 +1,7 @@
+from __future__ import annotations
+from calendar import c
+from cmath import exp
+import copy
 import enum
 from collections.abc import MutableMapping
 from configparser import ConfigParser, NoOptionError
@@ -6,16 +10,75 @@ import os
 import logging
 import re
 import json
+import hjson
 from typing import List
 
+class CfgValue:
 
-class ObjectValue:
+    def __init__(self) -> None:
+        pass
+
+    def __repr__(self) -> str:
+        return self.__str__()
+
+class UnitValue(CfgValue):
+
+    @classmethod
+    def s(cls, value):
+        return cls(value, "s")
+    
+    @classmethod
+    def m(cls, value):
+        return cls(value, "m")
+    
+    def __init__(self, value, unit) -> None:
+        self.value = value
+        self.unit = unit
+
+    def __str__(self) -> str:
+        return f"{self.value}{self.unit}"
+
+class BoolValue(CfgValue):
+    
+    def __init__(self, value) -> None:
+        self.value = value
+    
+    def __str__(self) -> str:
+        if self.value:
+            return "true"
+        else:
+            return "false"
+
+BoolValue.TRUE =  BoolValue(True)
+BoolValue.FALSE = BoolValue(False)
+
+class QString(CfgValue):
+    def __init__(self, value) -> None:
+        self.value = value
+    
+    def __str__(self):
+        return f'"{self.value}"'
+
+class ObjectValue(CfgValue):
 
     def __eq__(self, other):
         if isinstance(other, ObjectValue):
             return self._name == other._name and self._value == other._value
+        elif isinstance(other, str):
+            return self.__str__() == other
         return False
 
+    @staticmethod
+    def parse_value_str(str_val) -> dict:
+        try:
+            val_ = json.loads(str_val)
+        except JSONDecodeError:
+            try:
+                val_ = dict(hjson.loads(str_val))
+            except Exception as e:
+                raise ValueError(f"Error parsing '{str_val}' with json or hjson. Got: {e}")
+        return val_
+
     @classmethod
     def fromString(cls, str_val):
         """
@@ -24,7 +87,29 @@ class ObjectValue:
         Example: 'X{"key", "value"}'
         """
         obj_val = cls.parse_string(str_val)
-        return cls(name=obj_val[0], value=obj_val[1])
+        val = cls.parse_value_str(obj_val[1])
+        return cls(name=obj_val[0], value=val)
+
+    @classmethod
+    def from_args(cls, name, *args):
+        values = {}
+        if (len(args) % 2) != 0:
+            raise ValueError("expected even number of args")
+        for k, v in zip(args[0::2], args[1::2]):
+            values[k] = v
+        
+        return cls(name, values)
+
+    
+    def copy(self, *args):
+        _copy = ObjectValue(self._name, dict(self._value))
+        if (len(args) % 2) != 0:
+            raise ValueError("expected even number of args")
+        for k, v in zip(args[0::2], args[1::2]):
+            _copy[k] = v
+        
+        return _copy
+        
 
     def parse_string(str_val: str) -> List[str]:
         """
@@ -53,44 +138,54 @@ class ObjectValue:
         except:
             return False
 
-    def __init__(self, name="", value="{}") -> None:
+    def __init__(self, name="", value: str|dict={}) -> None:
         """
         string representation of the object. Value must be parsable to a valid json object {}
         It must containing the surrounding curly brackets. Value cannot be an array, only an object.
         Example: {key1: 42, key3: "Foo"}
         """
         self._name = name
-
-        self._value = value
+        if isinstance(value, str):
+            self._value = self.parse_value_str(value)
+        else:
+            self._value = value
 
     def __str__(self) -> str:
-        return f"{self._name}{self._value}"
+        _v = []
+        for k, v in self._value.items():
+            _v.append(f"{k}: {v}")
+        _v = "{"  + ", ".join(_v) + "}"
+        return f"{self._name}{_v}"
 
     @property
     def as_dict(self):
-        return json.loads(self._value)
+        return self._value
 
     @as_dict.setter
     def as_dict(self, val):
-        self._value = json.dumps(val)
+        self._value = val
 
     def add(self, key, value, override=True):
         """
         add key/value pair to self._value depending on override flag.
         If override False and value exist return false otherwise true. No error
         """
-        dict_val = json.loads(self._value)
         if override:
-            dict_val[key] = value
-            self.as_dict = dict_val
+            self._value[key] = value
             return True
         else:
             if key in self.as_dict:
                 return False
             else:
-                dict_val[key] = value
-                self.as_dict = dict_val
+                self._value[key] = value
                 return True
+    
+    def __setitem__(self, key, value):
+        self._value[key] = value
+    
+    def __getitem__(self, key):
+        return self._value[key]
+
 
 
 class OppParser(ConfigParser):
@@ -229,37 +324,64 @@ class OppParser(ConfigParser):
         return inc_file, lines[index]
 
 
-class OppConfigType(enum.Enum):
+class OppConfigType():
     """
     Set type on OppConfigFileBase to create read-only configurations if needed.
     """
 
-    READ_ONLY = 1
-    EDIT_LOCAL = 2
-    EXT_DEL_LOCAL = 3
+    # READ_ONLY = 1
+    # EDIT_LOCAL = 2
+    # EXT_DEL_LOCAL = 3
 
-    def __lt__(self, other):
-        if self.__class__ is other.__class__:
-            return self.value < other.value
-        elif type(other) is int:
-            return self.value < other
-        return NotImplemented
+    _read_only = 1 << 0
+    _edit_local = 1 << 1
+    _allow_deletion = 1 << 2
+    _edit_global = 1 << 3
+    _ext_allow_del_global = 1 << 4
 
-    def __gt__(self, other):
-        return self != other and other < self
 
-    def __eq__(self, other):
-        if self.__class__ is other.__class__:
-            return self.value == other.value
-        elif type(other) is int:
-            return self.value == other
-        return NotImplemented
+    def __init__(self, val) -> None:
+        self.val = val
+    
+    def __or__(self, v):
+        return self.val | v
+    
+    def __and__(self, v):
+        return self.val & v
+    
+    def __xor__(self, v):
+        return self.val ^ v
 
-    def __le__(self, other):
-        return self < other or self == other
+    @property
+    def read_only(self):
+        return (self & self._read_only) != 0
+    
+    @property
+    def edit_local(self):
+        return not self.read_only and ( self.edit_global or (self & self._edit_local) !=0)
+    
+    @property
+    def edit_global(self):
+        return not self.read_only and (self & self._edit_global) != 0
+    
+    @property
+    def allow_edit(self):
+        return self.edit_global or self.edit_local
+    
+    @property
+    def allow_delete(self):
+        return (self.edit_local or self.edit_global ) and (self & self._allow_deletion) != 0
+    
+
+
+OppConfigType.READ_ONLY = OppConfigType(OppConfigType._read_only)
+    
+OppConfigType.EDIT_LOCAL = OppConfigType(OppConfigType._edit_local)
+OppConfigType.EDIT_DEL_LOCAL = OppConfigType(OppConfigType.EDIT_LOCAL | OppConfigType._allow_deletion)
+
+OppConfigType.EDIT_GLOBAL = OppConfigType(OppConfigType.EDIT_LOCAL | OppConfigType._edit_global)
+OppConfigType.EDIT_DEL_GLOBAL = OppConfigType(OppConfigType.EDIT_GLOBAL | OppConfigType._allow_deletion )
 
-    def __ge__(self, other):
-        return self > other or self == other
 
 
 class OppConfigFileBase(MutableMapping):
@@ -316,7 +438,7 @@ class OppConfigFileBase(MutableMapping):
             is_parent=False,
     ):
         self._root: OppParser = root_cfg
-        self._cfg_type = cfg_type
+        self._cfg_type: OppConfigType = cfg_type
         self._sec = self._ensure_config_prefix(config_name)
         self._sec_raw = config_name
         self._base_dir = _base_dir
@@ -363,14 +485,23 @@ class OppConfigFileBase(MutableMapping):
     def section_hierarchy(self):
         return self._section_hierarchy
 
-    def resolve_path(self, key):
-        val = self.__getitem__(key)
+    def resolve_path(self, key, all:bool = False):
         p = re.compile('absFilePath\("(.*)"\)')
-        m = p.match(val)
-        if m:
-            return os.path.join(self.base_path, m.group(1))
+        def check(m):
+            m = p.match(val)
+            if m:
+                return os.path.join(self.base_path, m.group(1))
+            else:
+                return os.path.join(self.base_path, val.strip('"'))
+
+        if all :
+            ret = []
+            for val in self.get_all(key, with_section=False):
+                ret.append(check(val))
+            return ret
         else:
-            return os.path.join(self.base_path, val.strip('"'))
+            val = self.__getitem__(key)
+            return check(val)
 
     def read(self):
         self.read()
@@ -384,6 +515,7 @@ class OppConfigFileBase(MutableMapping):
         else:
             self._root.write(fp)
 
+
     @staticmethod
     def _ensure_config_prefix(val):
         """ All omnetpp configurations start with 'Config'. Add 'Config' if it is missing.  """
@@ -490,13 +622,13 @@ class OppConfigFileBase(MutableMapping):
         self._root.remove_option(self._sec, k)
 
     def __setitem__(self, k, v):
-        if self._cfg_type is OppConfigType.READ_ONLY:
+        if self._cfg_type.read_only:
             raise NotImplementedError("Cannot set value on read only config")
 
         if self._contains_local_(k):
             self._set_local(k, v)
         elif self._contained_by_parent(k):
-            if self._cfg_type <= OppConfigType.EDIT_LOCAL:
+            if not self._cfg_type.edit_global:
                 raise NotImplementedError("Cannot edit value of parent config")
             else:
                 for p in self._parent_cfg:
@@ -504,12 +636,14 @@ class OppConfigFileBase(MutableMapping):
                         p._set_local(k, v)
                         return
         else:
-            raise KeyError(f"key not found. Key: {k}")
+            # todo add option 
+            self._set_local(k, v)
+            # raise KeyError(f"key not found and object config does not allow addition of new keys. Key: {k}")
         if k == "extends":
             self.update_selection_hierarchy()
 
     def __delitem__(self, k):
-        if self._cfg_type.value < OppConfigType.EXT_DEL_LOCAL:
+        if not self._cfg_type.allow_delete:
             raise ValueError(
                 f"current object does not allow deletion. cfg_type={self._cfg_type}"
             )
@@ -566,19 +700,47 @@ class OppConfigFileBase(MutableMapping):
 
     def get(self, k):
         return self[k]
+    
+    def get_all(self, k, with_section: bool = True):
+        ret = {}
+        for _sec, _data in self._root._sections.items():
+            if k in _data:
+                ret[_sec] = _data[k]
+        
+        if with_section:
+            return ret
+        else:
+            return list(ret.values())
 
     def setdefault(self, k, default=...):
         if k in self:
             return self[k]
         else:
-            if self._cfg_type <= OppConfigType.EDIT_LOCAL:
+            if self._cfg_type.read_only:
                 raise NotImplementedError(
-                    "Cannot set value on READ_ONLY or EDIT_LOCAL config. Use EXT_DEL_LOCAL "
+                    "Object is read only"
                 )
             else:
                 self._set_local(k, default)
         return default
 
+def updateConfigFile(cfg: OppConfigFileBase, changes: dict, deepcopy: bool = True):
+    if deepcopy:
+        _c = copy.deepcopy(cfg)
+    else:
+        _c = cfg
+    if "extends" in changes:
+        _c["extends"] = changes["extends"]
+    for k, v in changes.items():
+        if k == "extends":
+            continue
+        if isinstance(v, CfgValue):
+            _c[k] = str(v)
+        else:
+            _c[k] = v
+
+    return _c
+
 
 class OppConfigFile(OppConfigFileBase):
     """
@@ -593,8 +755,3 @@ class OppConfigFile(OppConfigFileBase):
         super().__init__(root_cfg, config_name)
 
 
-if __name__ == "__main__":
-    OppParser.resolve_includes(
-        "/home/vm-sts/repos/crownet/crownet/simulations/mucFreiheitLte/omnetpp.ini",
-        output_path="/home/vm-sts/repos/crownet/crownet/simulations/mucFreiheitLte/omnetpp_merged.ini"
-    )
